#!/usr/bin/perl

# TODO:
# - sanity check to bail out early on files with line breaks in them?

use FindBin qw($Bin);
use lib $Bin;
use TreeHasher;

use File::Basename qw(basename dirname fileparse);
# This is still UNIX-specific. I do not care about VMS or Windows.
# Only using the tediours relative/absolute path resolution, I'll
# keep using slashes and '..' for the parent in verbatim.
use File::Spec::Functions qw(abs2rel rel2abs);
use Cwd qw(cwd realpath);
use File::Path qw(make_path);

use strict;
use warnings;

# Some bit-banging to reduce overhead from string matching,
# even runtime hash lookup.
my $mode_hash  = 1; # bit for hashing
my $mode_debug = 4; # bit for debugging
my %mode = (
	hash  => $mode_hash
,	debug => $mode_debug
,	verbose => 8
,	quiet   => 16
);

if(@ARGV < 3)
{
	print "Usage:\n\t$0 <modespec> <base> [directories to work on]\n\n"
	.	"The modespec is a comma-separated (no spaces) selection out of ("
	.	join(', ', sort keys %mode)."). "
	.	"Base is the base path of the treee being checksummed. Then you give a "
	.	"list of specific directories that should be worked on "
	.	"(can be identical to base)."
	.	"\n";
	exit(1);
}

# Actual mode of operation combination of the above.
my $mode = parse_mode(shift(@ARGV));
$verbose = 1;
$verbose = 2
	if $mode & $mode{verbose};
$verbose = 0
	if $mode & $mode{quiet};

TreeHasher::init(shift(@ARGV),1);
# Ensure that the timestamp of consecutive runs is unique by waiting one
# second. It's a hack, but since real work takes really long time,
# it should not hurt much.
sleep(1);

if($verbose)
{
	print msg('base',   $base);
	print msg('ckbase', $ckbase);
	print msg('mode',   $mode);
}

for my $dir (@ARGV)
{
	my $reldir = abs2rel(rel2abs($dir, $wrk), $base);
	# Only consider directories inside the base.
	die "$0: outsider: $reldir\n"
		if $reldir =~ m,^..(/|$),;
	# Do not confuse thyself by trying to check the checksums.
	next
		if $reldir =~ m:^$ckname/:o;
	work_dir($reldir);
}

sub parse_mode
{
	my $mode = 0;
	for(split(',', $_[0]))
	{
		if(defined $mode{$_})
		{
			# |= is experimental in Perl 5.22. I'm astounded that it's not core
			# since ages.
			$mode = $mode | $mode{$_};
		}
		else
		{
			die "$0: bad mode given: $_\n";
		}
	}
	return $mode;
}

# Merge exactly two arrays, preserving order, avoiding duplicates.
sub merged
{
	my @a = @{$_[0]};
	my %exist;
	@exist{@a} = ();
	push(@a, grep { not exists $exist{$_} } @{$_[1]});
	return \@a;
}


# The directory recursor. This does mess with cwd(). If you want
# you working directory to persist, ensure it yourself. This code
# has to cover cases with directories being deleted while it is
# traversing them. Keeping track of where we are is pain for no
# gain.
my $reclevel = 0;
sub work_dir
{
	if(++$reclevel > $reclimit)
	{
		die "$0: recursion limit of $reclimit reached\n"
	}
	my $reldir=shift;
	print msg('work('.$reclevel.')', $reldir)
		if $verbose;
	my $realdir = $base.'/'.$reldir;
	my $hashdir = $ckbase.'/'.$reldir;
	$realdir =~ s:/\.$::; # blabla/. = blabla
	$hashdir =~ s:/\.$::; # blabla/. = blabla
	if($verbose >= 2)
	{
		print msg('realdir', $realdir);
		print msg('hashdir', $hashdir);
	}
	# If the directory is static and a hash exists, for it, skip
	# any further hashing, also of stuff below it.
	my $lasthashtime = last_hashtime($reldir, $hash_alg);
	my $hashme = not (dir_is_static($reldir) and defined $lasthashtime);
	print msg('nohash/static', $reldir)
		if($verbose >=2 and not $hashme);
	die "$0: messed up, hash and real dirs the same\n"
		unless $realdir ne $hashdir;
 	# Really ensure that it's fully resolved now.
	if(defined (my $rrealdir = realpath($realdir)))
	{
		$realdir = $rrealdir;
	}
	else
	{
		# Might be recently deleted.
		print msg("unreal", $realdir);
		--$reclevel;
		return;
	}
	# I used to store/restore the working directory here.
	# This does not help anything as we can get lost anyway
	# When directories are deleted while operating.
	# Just always ensure full paths for chdir() and only
	# switching to directory to compute checksums therein.

	# To possibly clean up things, we need to dive into the union of hash and
	# real subdirectories. Non-existing subdirs on either side have to be
	# silently skipped.  Hm, what if something is a directory on one side and
	# a file on the other? Need to be aware of that.
	# 1. Descent into subdirectories.
	# Only in hashing mode, we acutally do need the potentially changed
	# real directories.
	my $subdirlist = ($mode & $mode_hash and $hashme)
	?	merged(dirs($realdir), dirs($hashdir))
	:	dirs($hashdir);
	for my $d (@{$subdirlist})
	{
		print msg('subdir('.$reclevel.')', $d)
			if $verbose >= 2;
		# Also jumping through hoops to avoid processing the actual checksum
		# root directory.
		my $relsubdir = $reldir eq '.' ? $d : $reldir.'/'.$d;
		next
			if($relsubdir eq $ckname);
		# Do all work in the sub directory.
		work_dir($relsubdir);
	}

	# 2. Compute checksums for files in this directory, if necessary.
	# This creates a hash sum file at the same level as the hash directory,
	# not inside it.
	# Failure to change into real directory is accepted since it might
	# be removed by users once we arrive there. Generally, we are gracious
	# on the real data side with errors.
	if($mode & $mode_hash and $hashme and -d $realdir)
	{
		if(dir_is_fresher($realdir, $lasthashtime))
		{
			print msg('hashing', $realdir)
				if $verbose;
			if(chdir($realdir))
			{
				my $hashfile = hashfile_name($hashdir, $hash_alg, $timestamp);
				print msg('hashfile', $hashfile)
					if $verbose >= 2;
				unless($mode & $mode_debug)
				{
					make_path(dirname($hashdir));
					write_hashes($hashfile, 0, @{files()});
				}
			}
			else
			{
				print msg('skip bad dir', $realdir)
					if $verbose;
			}
		}
		else
		{
			print msg('skip hashing', $realdir)
				if $verbose;
		}
	}

	# This is a relic from the time where we merged hash files. But it
	# may still be sensible once an option of cleaning up old hashes
	# enters.
	# 4. Remove hash directory if empty. NEVER TOUCH THE REAL DATA DIRECTORY!
	if(dir_is_empty($hashdir))
	{
		if($mode & $mode_debug)
		{
			print msg('remove', $hashdir)
				if $verbose;
		}
		else
		{
			if(rmdir($hashdir))
			{
				print msg('removed', $hashdir)
					if $verbose;
			}
		}
	}
	--$reclevel;
}
