#!/usr/bin/perl

# TODO: proper FTP support that can distinguish files from directories,
# as lynx -dump doesn't append a / for me:-/
# So I'll resort to wget after all and parse the generated HTML.
# For HTTP I'll probably do that, too, but lynx is still convenient
# to strip out all irrelevant stuff. It's still a different matter parsing
# any HTML from whatever source and the HTML generated by wget itself from
# a ftp listing.
# Btw, parsing curl output with Perl is simpler than either calling a FTP client
# or managing the stuff with Net::FTP myself. That way, I don't even have to
# deconstruct the urls.

use strict;

use FindBin qw($Bin);
use lib "$Bin/../lib/perl";

use Time::HiRes qw(sleep);

use Config::Param;

$ENV{LANG} = 'C'; # No surprises from spawned programs.

my $p = Config::Param::get
({
	tagline=>'extract URLs to relevant files from some index page'
,	usage=>"$0 [params] <url to index page>"
,	info=>'This uses the lynx browser for parsing the pages via lynx -dump.'
},[
	'recursive', 0, 'r', 'collect links recursively up to this depth (always with a limit!)'
,	'recurse-root', '', '', 'enforce a specific URL to consider as root for recursion (in case the primary URL redirects)'
,	'relevant-link',[],'L','patterns that a link should match before processing for file and dir part'
,	'relevant', ['.*\.(tgz|tar|zip|diff)(\..+|)*$'], 'R', 'what relevant means (list of regexes that need to match on the file name, AND)'
,	'irrelevant', [], 'E', 'file names matching one of those are irrelevant'
,	'relevant-dir', [], '', 'list of regexes for sub-directories to follow (if non-empty, at least one of the regexes needs to match, OR)'
,	'irrelevant-dir', [], '', 'list of regexes for sub-directories to ignore (overrules relevant-dir)'
,	'sleep', 0.1, 's', 'time in seconds to sleep between requests (some servers do not like the load)'
,	'debug',0,'','debugging output'
,	'link-dirfile', '', '', 'regex matching for dir and file parts (\$1 and \$2); If that does not match, the usual processing takes place.'
,	'keep-path','',''
	,	'Keep paths relative to the given value and omit files if it the URL does not start with this; the given value is blindly removed from the URLs and the rest is used as local path. Make sure you include the trailing / if you want to have things below a certain directory.'
,	'dircache',['.'],''
	,	'regex list (OR) to match the full path of a directory below the root URL, to be put into the dircache file / not being searched again if found in the same'
,	'dircache-file',undef,''
	,	'file to read / write directory cache from / to'
]);

my $index_url = shift @ARGV;
my @relevant_ex;
for(@{$p->{relevant}})
{
	debug("RELEVANT: $_\n");
	push(@relevant_ex, qr/$_/);
}
my @irrelevant_ex;
for(@{$p->{irrelevant}})
{
	debug("IRRELEVANT: $_\n");
	push(@irrelevant_ex, qr/$_/);
}
my @reldir_ex;
for(@{$p->{'relevant-dir'}})
{
	debug("RELEVANT-DIR: $_\n");
	push(@reldir_ex, qr/$_/);
}
my @rellink_ex;
for(@{$p->{'relevant-link'}})
{
	debug("RELEVANT-LINK: $_\n");
	push(@rellink_ex, qr/$_/);
}
my @irreldir_ex;
for(@{$p->{'irrelevant-dir'}})
{
	debug("IRRELEVANT-DIR: $_\n");
	push(@irreldir_ex, qr/$_/);
}
my $linkdirfile_ex;
if($p->{'link-dirfile'} ne '')
{
	$linkdirfile_ex = qr/$p->{'link-dirfile'}/;
}
my @dircache_ex;
for(@{$p->{dircache}})
{
	debug("DIRCACHE: $_\n");
	push(@dircache_ex, qr/$_/);
}
my $use_dircache = 0;
my $new_dirs = 0;
my @dircache;
if(defined $p->{'dircache-file'})
{
	$use_dircache = 1;
	if(open(my $fh, '<', $p->{'dircache-file'}))
	{
		while(<$fh>)
		{
			chomp;
			push(@dircache, $_);
		}
		close($fh);
	}
}

$index_url = resolve_redirect($index_url);

my ($globalroot) = $p->{'recurse-root'} ne ''
?	($p->{'recurse-root'})
:	url_filesplit($index_url);
debug("GLOBAL ROOT: $globalroot\n");

extract_urls($index_url, 0);

if($use_dircache and $new_dirs)
{
	if(open(my $fh, '>>', $p->{'dircache-file'}))
	{
		for(my $i=@dircache-$new_dirs; $i<@dircache; ++$i)
		{
			print $fh $dircache[$i]."\n";
		}
		close($fh);
	}
	else{ die "$0: cannot update dircache file $p->{'dircache-file'}: $!\n"; }
}


exit;

# This is getting ridiculous. I should use one of them HTTP Perl Modules.
# But then, this is UNIX, employing a standalone tool for redirection.
# Use case: A mirror load balancer that redirects to some mirror.
# I need to contact the mirror directly so that recursive directory
# matching works.
sub resolve_redirect
{
	my $url = shift;
	open(my $redir, '-|', $Bin.'/resolve-redirect', $url)
		or die "Cannot run redirector!\n";
	my $nurl = <$redir>;
	chomp($nurl);
	close($redir);
	debug("resolved $url to $nurl\n");
	sleep($p->{sleep});
	return $nurl;
}

sub extract_urls
{
	my $index_url = shift;
	my $depth     = shift;
	my ($root, $indexpage) = url_filesplit($index_url);
	$root = $globalroot
		if $depth == 0;
	debug("ROOT: $root\n");
	my @urls;
	# It might work that we extract some links, but not all.
	# Avoid putting into directory cache if we did not get all.
	my $ret = get_links($index_url, \@urls);
	for my $url (@urls)
	{
		next unless is_relevant_link($url);
		my ($dir, $file) = url_filesplit($url);
		debug("DIR: $dir; FILE: $file\n");
		if( $p->{recursive}
		    and $file eq '.'
		    and substr($dir,0,length($root)+1) eq $root.'/' )
		{
			# directory below $root
			my $subdir  = substr($dir, length($root)+1);
			my $gsubdir = substr($dir, length($globalroot)+1);
			# directory below $globalroot
			debug("RECURSE? $depth < $p->{recursive}? SUB: $subdir GSUB: $gsubdir\n");
			if( $depth < $p->{recursive} and is_relevant_dir($subdir)
				and not_in_dircache($gsubdir) )
			{
				debug("RECURSE: $url\n");
				# Don't be abusive with recursive requests.
				sleep($p->{sleep});
				if(extract_urls($url, $depth+1))
				{
					# Only add to cache after one successful run.
					add_to_dircache($gsubdir);
				}
			}
			else
			{
				debug("SKIP: $url\n");
			}	
		}
		elsif(defined $file and is_relevant($file))
		{
			my $localfile = $file;
			if($p->{'keep-path'} ne '')
			{
				my $localdir = url_decode($dir);
				next unless	$localdir =~ s:^$p->{'keep-path'}::;
				$localfile = $localdir;
				$localfile .= '/' if $localfile;
				$localfile .= $file;
			}
			print $localfile."\t".$url."\n";
		}
	}
	return $ret;
}

# Return 1 if any of the provided regexes matches the first argument.
# In list context, also returns the regex that did match.
sub list_match_or
{
	my $cand = shift;
	my $debug_msg = shift;
	for(@_)
	{
		if($cand =~ $_)
		{
			debug($debug_msg." '$cand' =~ m|$_|\n");
			return 1;
		}
	}
	return 0;
}

# Return 1 if all of the provided regexes match the first argument.
sub list_match_and
{
	my $cand = shift;
	my $debug_msg = shift;
	for(@_)
	{
		unless($cand =~ $_)
		{
			debug($debug_msg." not '$cand' =~ m|$_|\n");
			return 0;
		}
	}
	return 1;
}

sub is_relevant
{
	my $file = shift;
	return 0
		if list_match_or($file, "IRRELEVANT!", @irrelevant_ex);
	return 1
		if list_match_and($file, "NOT RELEVANT!", @relevant_ex);
	return 0;
}

sub is_relevant_dir
{
	my $subdir = shift;
	for(my $i=0; $i<@irreldir_ex; ++$i)
	{
		if($subdir =~ $irreldir_ex[$i])
		{
			debug("IRRELEVANT! '$subdir' =~ /$p->{'irrelevant-dir'}[$i]/\n");
			return 0;
		}
	}
	for(my $i=0; $i<@reldir_ex; ++$i)
	{
		if($subdir =~ $reldir_ex[$i])
		{
			debug("RELEVANT! '$subdir' =~ /$p->{'relevant-dir'}[$i]/\n");
			return 1;
		}
	}
	debug("RELEVANT-STILL-THERE: $subdir\n");
	return @reldir_ex ? 0 : 1;
}

sub is_relevant_link
{
	my $link = shift;
	for(@rellink_ex)
	{
		return 1 if $link =~ $_;
	}
	return @rellink_ex ? 0 : 1;
}

# just get all direkt links, filter later
sub get_links
{
	my $index_url = shift;
	my $urls = shift;
	my $ret = 0; # bad
	$ret = get_links_http($index_url, $urls)
		if($index_url =~ m,^http,);
	$ret = get_links_ftp($index_url, $urls)
		if($index_url =~ m,^ftp,);

	return $ret; # true if we think we got everything
}

sub get_links_http
{
	my $index_url = shift;
	my $urls = shift;
	my $refmode;

	open( my $lynx, '-|', 'lynx'
	,	'-connect_timeout=23', '-read_timeout=23'
	,	'-dump', '-listonly', $index_url )
		or die "Cannot start lynx!\n";
	while(<$lynx>)
	{
		debug("LYNX: $_");
		chomp;
		# Make sure to collect only the reference list at the very end.
		if(/^References$/)
		{
			@{$urls} = ();
			$refmode = 1;
			next;
		}
		next unless $refmode;
		if(/^\s*\d+\.\s+(.*)$/)
		{
			push(@{$urls}, $1);
			debug("pushed: $1\n");
		}
		
	}
	close($lynx);
	return $? == 0; # exit status of lynx
}

sub get_links_ftp
{
	my $index_url = shift;
	my $urls = shift;

	open( my $curl, '-|', 'curl'
	,	'--max-time', 60, '--connect-timeout', '23', '-s'
	,	$index_url )
		or die "Cannot start curl!\n";
	my $prefix = $index_url;
	$prefix .= '/' unless $prefix =~ m:/$:;
	while(<$curl>)
	{
		debug("CURL: $_");
		chomp;
		#-rw-r--r--    1 3568     800      11771056 Mar 13 23:26 mesa-10.5.1.tar.gz
		#drwxrwxr-x    2 2549     800          4096 Mar 05  2014 10.1
		# I'll ignore files with spaces in them!
		if(/^dr.*\s+(\S+)$/)
		{
			push(@{$urls}, "$prefix$1/");
		}
		elsif(/^-r.*\s+(\S+)$/)
		{
			push(@{$urls}, "$prefix$1");
		}
	}
	close($curl);
	return $? == 0; # exit status of curl
}

# Extract file name out of URL.
sub url_filesplit
{
	my $url = url_decode(shift);
	my $file;
	my $dir;

	debug("splitting URL: $url ".(defined $linkdirfile_ex ? "m|$linkdirfile_ex|" : "<verbatim>")."\n");
	if( defined $linkdirfile_ex
	   and $url =~ $linkdirfile_ex )
	{
		$dir = $1;
		$file = $2;
		debug("matched dir and file with user expression\n");
	}
	else
	{
		# Some smart folks attach stuff like #md5=... ignore any supposed
		# anchor crap like that.
		$url =~ s:#.*$::;
		if($url =~ m,^([^\?]+)\?(.*)$,)
		{
			my $urlpart = $1;
			my $parpart = $2;
			$dir = $urlpart;
			$dir =~ s,/*[^/]*$,,;
			# TODO: parse that, looking out for file= or filename= etc.
			debug("not parsing file name: $url\n");
		}
		else
		{
			debug("bog-standard split\n");
			$dir = $url;
			$dir =~ s,/*[^/]*$,,;
			$file = $url =~ m,/$,
			?	'.'
			:	pop(@{[split('/', $url)]});
		}
	}
	if($file =~ m,\t,)
	{
		debug("No, a TAB in a file name?! Dropping that: $file\n");
		$file = undef;
	}
	# index.html is the directory itself
	$file = '.'
		if $file =~ /^index.s?html$/;
	return ($dir, $file);
}

# I just felt like including that one here, could have properly
# used the module I originally put that one into. Could have used
# a proper huge CPAN/Perl module. Didn't feel like it. At least I know
# that code.
sub url_decode
{
	use bytes; # For hacking unicode together.
	my $urlpart = shift;

	my $decoded;
	# WRONG! This is only for arguments (after ?).
	# Need to adapt that once I support them.
	# $urlpart =~ s/\+/ /g; # Empty spaces...
	# C-ish parsing... I want to be explicit and clear here.
	my $oldpos = 0;
	my $length = length($urlpart);
	while((my $pos = index($urlpart, '%', $oldpos)) >= 0)
	{ # Now copy stuff up to %xx, decode the hex sequence.
		$decoded .= substr($urlpart, $oldpos, $pos-$oldpos);
		if($length-$pos < 3)
		{ # We're cut! no full %xx triplet left!
		        $oldpos = $length;
		        last;
		}
		# Append the hex-decoded byte.
		$decoded .= chr(hex(substr($urlpart, $pos+1, 2)));
		$oldpos = $pos+3; # From %xxA to A, skip 3 characters.
	}
	# Take the rest.
	$decoded .= $oldpos < $length ? substr($urlpart, $oldpos) : '';
	return $decoded;
}

sub not_in_dircache
{
	my $dir = shift;
	return 1
		unless $use_dircache;
	if(grep { $_ eq $dir } @dircache)
	{
		debug("DIRCACHE-CACHED: $dir\n");
		return 0;
	}
	else
	{
		return 1;
	}
}

# Return value indicates if dir is (to be) cached or not.
sub add_to_dircache
{
	my $dir = shift;
        return 1
                unless $use_dircache;
	debug("DIRCACHE-ADD: $dir\n");
        if(grep { $_ eq $dir } @dircache)
        {
		debug("DIRCACHE-PRESENT: $dir\n");      
                return 1;
        } elsif(list_match_or($dir, 'CACHEWORTHY!', @dircache_ex))
        {      
                debug("DIRCACHE-NEW: $dir\n");
                push(@dircache, $dir);
                ++$new_dirs;
                return 1;
        }
	else
	{
		return 0;
	}
}

sub debug
{
	print STDERR @_
		if($p->{debug});
}
